// File:  timing.v
// Date:  01/01/2005
// Name:  Eric Crabill
//
// This is an implementation of a video timing
// controller for a 400x320@72fps non-interlaced
// raster display.  It is derived from the standard
// 800x600@72fps VESA mode.  It is hardly optimal;
// a quick kludge to simplify a previously written
// module.  The clock input must be 50 MHz.  During
// active video, when blank is not asserted, hcount
// and vcount will indicate the beam location, with
// coordinates on hcount from 0 to 399, and vcount
// from 0 to 319.

`timescale 1 ns / 1 ps

module timing (
                hcount,
                hsync,
                vcount,
                vsync,
                blank,
                clk,
                rst
                );

  output  [8:0] hcount;
  output        hsync;
  output  [8:0] vcount;
  output        vsync;
  output        blank;
  input         clk;
  input         rst;


  //******************************************************************//
  // This is the 12-bit horizontal counter.                           // 
  //******************************************************************//
 
 
  reg    [11:0] hpos_cnt;
  wire          hpos_clr;

  always @(posedge clk or posedge rst)
  begin : hcounter
    if (rst) hpos_cnt <= 12'h000;
    else
    begin
      if (hpos_clr) hpos_cnt <= 12'h000;
      else hpos_cnt <= hpos_cnt + 12'h001;
    end
  end


  //******************************************************************//
  // This is the 12-bit vertical counter.                             // 
  //******************************************************************//


  reg    [11:0] vpos_cnt;
  wire          vpos_clr;
  wire          vpos_ena;

  always @(posedge clk or posedge rst)
  begin : vcounter
    if (rst) vpos_cnt <= 12'h000;
    else
    begin
      if (vpos_clr) vpos_cnt <= 12'h000;
      else if (vpos_ena) vpos_cnt <= vpos_cnt + 12'h001;
    end
  end


  //******************************************************************//
  // This is the counter control logic.                               // 
  //******************************************************************//
 

  assign hpos_clr = (hpos_cnt >= 12'h40f);
  assign vpos_clr = (vpos_cnt >= 12'h299) & hpos_clr;
  assign vpos_ena = hpos_clr;


  //******************************************************************//
  // Horizontal timing logic.                                         //
  //******************************************************************//


  reg    [11:0] reg_hcount;
  reg           reg_hsync;
  reg           reg_hblnk;
 
  always @(posedge clk or posedge rst)
  begin : houtputs
    if (rst)
    begin
      reg_hcount <= 12'h000;
      reg_hsync <= 1'b1;
      reg_hblnk <= 1'b1;
    end
    else
    begin
      reg_hcount <= hpos_cnt - 12'h180;
      reg_hsync  <= (hpos_cnt <= 12'h077);
      reg_hblnk  <= (hpos_cnt <= 12'h17F) | (hpos_cnt > 12'h30F);
    end
  end

  assign hcount = reg_hcount[8:0];
  assign hsync = reg_hsync;


  //******************************************************************//
  // Vertical timing logic.                                           //
  //******************************************************************//
 

  reg    [11:0] reg_vcount;
  reg           reg_vsync;
  reg           reg_vblnk;
 
  always @(posedge clk or posedge rst)
  begin : voutputs
    if (rst)
    begin
      reg_vcount <= 12'h000;
      reg_vsync <= 1'b1;
      reg_vblnk <= 1'b1;
    end
    else
    begin
      reg_vcount <= vpos_cnt - 12'h0a9;
      reg_vsync  <= (vpos_cnt <= 12'h005);
      reg_vblnk  <= (vpos_cnt <= 12'h0a8) | (vpos_cnt > 12'h1e8);
    end
  end
 
  assign vcount = reg_vcount[8:0];
  assign vsync = reg_vsync;
 

  //******************************************************************//
  // Generate composite blank signal.                                 //
  //******************************************************************//


  assign blank = reg_vblnk | reg_hblnk;


  //******************************************************************//
  //                                                                  //
  //******************************************************************//


endmodule
